@@ -12,27 +12,27 @@ L{Element},  L{Project}. Also, two additional classes are defined for the
 association relationships between the former: L{TagElement} and L{TagProject}.
 """
 
+from version import *
+
 import os
 import sys
+import warnings
 
-import sqlalchemy
-
-from sqlalchemy import (Column, MetaData, Table, Index,
-                        ForeignKeyConstraint, PrimaryKeyConstraint, UniqueConstraint,
-                        Boolean, Integer, String, Text,
-                        engine_from_config, func, event)
+from sqlalchemy import inspect
+from sqlalchemy.engine import create_engine, Engine
+from sqlalchemy.events import event
 from sqlalchemy.ext.associationproxy import association_proxy
 from sqlalchemy.ext.hybrid import hybrid_property
 from sqlalchemy.orm import mapper, relationship, scoped_session, sessionmaker
 from sqlalchemy.orm.collections import attribute_mapped_collection
 from sqlalchemy.orm.exc import NoResultFound
-from sqlalchemy.orm.session import Session
-from sqlalchemy.sql.expression import literal
-
-try:
-    import json
-except: # pragma: no cover
-    import simplejson as json
+from sqlalchemy.orm.session import object_session
+from sqlalchemy.orm.util import identity_key
+from sqlalchemy.schema import Column, MetaData, Table
+from sqlalchemy.schema import ForeignKeyConstraint, Index, PrimaryKeyConstraint, UniqueConstraint
+from sqlalchemy.sql import functions
+from sqlalchemy.sql.expression import literal, literal_column
+from sqlalchemy.types import Boolean, DateTime, Integer, String, Text
 
 #: Metadata object for this model
 metadata = MetaData()
@@ -58,7 +58,6 @@ def init(session_maker, **kwargs):
     setup_mappers(**kwargs)
     setup_events(session_maker, **kwargs)
 
-
 def setup_tables():
     """
     Define the tables, columns, keys and constraints of the DB.
@@ -83,6 +82,16 @@ def setup_tables():
         postgresql_where=tables['element'].c.parent_id == None, unique=True
     )
     
+    tables['status_history'] =  Table('status_history', metadata,
+        Column('id',            Integer,     nullable=False),
+        Column('element_id',    Integer,     nullable=False),
+        Column('timestamp',     DateTime,    nullable=False, default=functions.now()),
+        Column('status',        String(20),  nullable=False),
+        PrimaryKeyConstraint('id'),
+        ForeignKeyConstraint(['element_id'], ['element.id'], ondelete='CASCADE'),
+    )
+    Index('status_history_ik_order', tables['status_history'].c.element_id, tables['status_history'].c.timestamp)
+    
     tables['project'] = Table('project', metadata,
         Column('id',   Integer,    nullable=False),
         Column('name', String(20), nullable=False),
@@ -91,18 +100,18 @@ def setup_tables():
     )
     
     tables['tag_element'] = Table('tag_element', metadata,
-        Column('tag',        String(20), nullable=False),
+        Column('name',       String(20), nullable=False),
         Column('element_id', Integer,    nullable=False),
         Column('value',      Text(),     nullable=True),
-        PrimaryKeyConstraint('element_id', 'tag'),
+        PrimaryKeyConstraint('element_id', 'name'),
         ForeignKeyConstraint(['element_id'], ['element.id'], ondelete='CASCADE'),
     )
     
     tables['tag_project'] = Table('tag_project', metadata,
-        Column('tag',        String(20), nullable=False),
+        Column('name',       String(20), nullable=False),
         Column('project_id', Integer,    nullable=False),
         Column('value',      Text(),     nullable=True),
-        PrimaryKeyConstraint('project_id', 'tag'),
+        PrimaryKeyConstraint('project_id', 'name'),
         ForeignKeyConstraint(['project_id'], ['project.id'], ondelete='CASCADE'),
     )
 
@@ -123,41 +132,51 @@ def setup_mappers(**kwargs):
     global mappers
     
     mapping = {
-        'element'     : Element,
-        'project'     : Project,
-        'tag_element' : TagElement,
-        'tag_project' : TagProject,
+        'element'        : Element,
+        'status_history' : StatusHistory,
+        'project'        : Project,
+        'tag_element'    : TagElement,
+        'tag_project'    : TagProject,
     }
     mapping.update(kwargs.get('extra_mapping', dict()))
     
-    assert issubclass(mapping['element'],     Element)
-    assert issubclass(mapping['project'],     Project)
-    assert issubclass(mapping['tag_element'], TagElement)
-    assert issubclass(mapping['tag_project'], TagProject)
+    assert issubclass(mapping['element'],        Element)
+    assert issubclass(mapping['status_history'], StatusHistory)
+    assert issubclass(mapping['project'],        Project)
+    assert issubclass(mapping['tag_element'],    TagElement)
+    assert issubclass(mapping['tag_project'],    TagProject)
     
     properties = {}
     properties['element'] = {
-        '_id'           : tables['element'].c.id,
-        '_parent_id'    : tables['element'].c.parent_id,
-        '_project_id'   : tables['element'].c.project_id,
-        '_status'       : tables['element'].c.status,
-        # FIXME: Add project_id to parent and children properties when #1401 is fixed
-        '_children'     : relationship(mapping['element'],     back_populates='_parent',   collection_class=set,
-                                       primaryjoin = tables['element'].c.parent_id == tables['element'].c.id,
-                                       cascade='all', passive_deletes=True),
-        '_parent'       : relationship(mapping['element'],     back_populates='_children', collection_class=set,
-                                       primaryjoin = tables['element'].c.parent_id == tables['element'].c.id,
-                                       remote_side = [ tables['element'].c.id ]),
-        '_project'      : relationship(mapping['project'],     back_populates='elements',  collection_class=set),
-        '_tag'          : relationship(mapping['tag_element'], collection_class=attribute_mapped_collection('tag'),
-                                       cascade='all, delete-orphan', passive_deletes=True),
+        '_id'            : tables['element'].c.id,
+        '_parent_id'     : tables['element'].c.parent_id,
+        '_project_id'    : tables['element'].c.project_id,
+        '_status'        : tables['element'].c.status,
+        '_children'      : relationship(mapping['element'],     back_populates='_parent',   collection_class=set,
+                                        primaryjoin = tables['element'].c.parent_id == tables['element'].c.id,
+                                        cascade='all', passive_deletes=True),
+        '_parent'        : relationship(mapping['element'],     back_populates='_children', collection_class=set,
+                                        primaryjoin = tables['element'].c.parent_id == tables['element'].c.id,
+                                        remote_side = [ tables['element'].c.id ]),
+        '_project'       : relationship(mapping['project'],     back_populates='elements',  collection_class=set),
+        'status_history' : relationship(mapping['status_history'], back_populates='_element',
+                                        order_by=tables['status_history'].c.id, viewonly=True),
+        'tags'           : relationship(mapping['tag_element'], collection_class=attribute_mapped_collection('name'),
+                                        cascade='all, delete-orphan', passive_deletes=True),
+    }
+    properties['status_history'] = {
+        '_id'         : tables['status_history'].c.id,
+        '_element_id' : tables['status_history'].c.element_id,
+        '_timestamp'  : tables['status_history'].c.timestamp,
+        '_status'     : tables['status_history'].c.status,
+        '_element'     : relationship(mapping['element'], back_populates='status_history'),
     }
     properties['project'] = {
         '_id'      : tables['project'].c.id,
         'elements' : relationship(mapping['element'],     back_populates='_project', collection_class=set,
                                   primaryjoin = tables['element'].c.project_id == tables['project'].c.id,
                                   cascade='all', passive_deletes=True),
-        '_tag'     : relationship(mapping['tag_project'], collection_class=attribute_mapped_collection('tag'),
+        'tags'     : relationship(mapping['tag_project'], collection_class=attribute_mapped_collection('name'),
                                   cascade='all, delete-orphan', passive_deletes=True),
     }
     properties['tag_element'] = {}
@@ -183,25 +202,34 @@ def setup_mappers(**kwargs):
     Association proxy to access its tags and retrieve their corresponding value.
     Example: instance.tag['name'] = 'value'
     """
-    Element.tag = association_proxy('_tag', 'value', creator=lambda tag, value: mapping['tag_element'](tag=tag, value=value))
-    Project.tag = association_proxy('_tag', 'value', creator=lambda tag, value: mapping['tag_project'](tag=tag, value=value))
-
+    Element.tag = association_proxy('tags', 'value', creator=lambda name, value: mapping['tag_element'](name=name, value=value))
+    Project.tag = association_proxy('tags', 'value', creator=lambda name, value: mapping['tag_project'](name=name, value=value))
 
 def setup_events(session_maker, **kwargs):
     """
     Define the events of the model.
     """
     mapping = {
-        'element'     : Element,
-        'project'     : Project,
-        'tag_element' : TagElement,
-        'tag_project' : TagProject,
+        'element'        : Element,
+        'status_history' : StatusHistory,
+        'project'        : Project,
+        'tag_element'    : TagElement,
+        'tag_project'    : TagProject,
     }
     mapping.update(kwargs.get('extra_mapping', dict()))
     
     event.listen(mapping['element']._children, 'append', mapping['element']._children_added)
     event.listen(mapping['element']._children, 'remove', mapping['element']._children_removed)
     event.listen(session_maker, 'before_flush', _session_before_flush)
+    event.listen(Engine, "begin", _sqlite_begin)
+
+def _sqlite_begin(conn):
+    if conn.engine.name == 'sqlite':
+        # Foreign keys are NOT enabled by default... WTF!
+        conn.execute("PRAGMA foreign_keys = ON")
+        # Force a single active transaction on a sqlite database.
+        # This is needed to emulate FOR UPDATE locks :(
+        conn.execute("BEGIN EXCLUSIVE")
 
 def _session_before_flush(session, flush_context, instances):
     """
@@ -212,6 +240,29 @@ def _session_before_flush(session, flush_context, instances):
         if isinstance(instance, Element):
             if instance.parent:
                 instance.parent.children.remove(instance)
+    
+    for instance in session.new:
+        if isinstance(instance, Element):
+            session.add(StatusHistory(
+                _element = instance,
+                _status  = instance.status
+            ))
+    
+    for instance in session.dirty:
+        if isinstance(instance, Element):
+            if inspect(instance).attrs['_status'].history[2]:
+                session.add(StatusHistory(
+                    _element = instance,
+                    _status  = instance.status
+                ))
+
+class CTENotSupported(UserWarning):
+    """
+    User-defined exception to warn the user when calling a method
+    that uses Common Table Expressions (CTE) to perform its work.
+    In most cases, there is an alternative slow path of code.
+    """
+    pass
 
 class ORM_Base(object):
     """
@@ -231,7 +282,6 @@ class ORM_Base(object):
         for name, value in kwargs.iteritems():
             setattr(self, name, value)
 
-
 class Element(ORM_Base):
     """
     Mapping class for the table «element».
@@ -441,33 +491,51 @@ class Element(ORM_Base):
             if self.parent.status != new_status:
                 self.parent.status = new_status 
     
-    @classmethod
-    def query_tags(cls, session, tags={}):
+    @property
+    def ancestors(self):
         """
-        Overriden query method to apply tag-based custom filtering, on top of
-        common equality filter.
+        Retrieve all the ancestors of this node.
+        If the node has no parents, return an empty list.
+        Else, start retrieving them from the identity map and, when not there,
+        fetch the rest from the database using a CTE.
         
-        @param session: The database session in which to execute the query
-        @type session: Session
-        @param tags: Tag names and values to apply as a filter
-        @type tags: dict
-        @return: A query selecting Element instances, filtered by tags
-        @rtype: Query<L{Element}>
+        @return: A list of all the ancestors of this node, ordered by proximity.
+        @rtype: list<L{Element}>
         """
-        
-        q = session.query(cls)
-        if tags:
-            crit = literal(False)
-            for tag, value in tags.iteritems():
-                if value == True: 
-                    crit |= (tables['tag_element'].c.tag == tag)
-                elif value == None or value != False :
-                    crit |= (tables['tag_element'].c.tag == tag) & (tables['tag_element'].c.value == value)
-            q = q.join(tables['tag_element']) \
-                 .group_by(*[c for c in tables['element'].columns]) \
-                 .having(func.count(tables['element'].c.id) == len(tags)) \
-                 .filter(crit)
-        return q.from_self()
+        cls = self.__class__
+        session = object_session(self)
+        # Must flush to get the parent_id from the database
+        session.flush()
+        if self._parent_id is None:
+            # End of the recursion. This Element has no parent.
+            return []
+        else:
+            # Best case available: retrieve the parent from the identity map
+            key = identity_key(cls, self._parent_id)
+            parent = session.identity_map.get(key, None)
+            if parent:
+                # Parent found in identity map, recurse and return.
+                parents = [parent]
+                parents.extend(parent.ancestors)
+                return parents
+            # Parent NOT found in identity map. Must use CTE
+            if session.bind.name != 'postgresql':
+                # If not using PostgreSQL, warn the user and use the non-optimized method.
+                warnings.warn('CTE are only supported on PostgreSQL. Using slower technique for "ancestor" method.', CTENotSupported)
+                parents = [self.parent]
+                parents.extend(self.parent.ancestors)
+                return parents
+            # Use a CTE to retrieve ALL ancestors
+            l0 = literal_column('0').label('level')
+            q_base = session.query(cls, l0).filter_by(
+                         id = self._parent_id
+                     ).cte(recursive = True)
+            l1 = literal_column('level + 1').label('level')
+            q_rec = session.query(cls, l1).filter(
+                        q_base.c.parent_id == cls.id
+                    )
+            q_cte = q_base.union_all(q_rec)
+            return session.query(cls).select_from(q_cte).order_by(q_cte.c.level)
     
     def __repr__(self):
         """
@@ -476,7 +544,15 @@ class Element(ORM_Base):
         @return: A descriptive string containing most of this instance fields
         @rtype: str
         """
-        return u"<Element %s: name '%s', parent_id '%s', project_id '%s', status '%s', forced_status '%s'>" % (self.id, self.name, self.parent_id, self.project_id, self.status, self.forced_status)
+        return u"%s(id=%s, name=%s, parent_id=%s, project_id=%s, status=%s, forced_status=%s)" % (
+            self.__class__.__name__,
+            repr(self.id),
+            repr(self.name),
+            repr(self.parent_id),
+            repr(self.project_id),
+            repr(self.status),
+            repr(self.forced_status),
+        )
     
     def __str__(self):
         """
@@ -487,6 +563,83 @@ class Element(ORM_Base):
         """
         return str(self.name)
 
+class StatusHistory(ORM_Base):
+    """
+    Mapping class for the table «status_history».
+    """
+    
+    @hybrid_property
+    def id(self):
+        """
+        Read only accessor to prevent setting this field.
+        
+        @return: Surrogate primary key
+        @rtype: int
+        """
+        return self._id
+    
+    @hybrid_property
+    def element_id(self):
+        """
+        Read only accessor to prevent setting this field.
+        
+        @return: Surrogate primary key
+        @rtype: int
+        """
+        return self._element_id
+    
+    @hybrid_property
+    def element(self):
+        """
+        The related Element of this instance.
+        
+        @return: The related Element
+        @rtype: L{Element}
+        """
+        return self._element
+    
+    @hybrid_property
+    def timestamp(self):
+        """
+        Read only accessor to prevent setting this field.
+        
+        @return: The time this status was updated.
+        @rtype: datetime
+        """
+        return self._timestamp
+    
+    @hybrid_property
+    def status(self):
+        """
+        The status of this instance
+        
+        @return: status
+        @rtype: str
+        """
+        return self._status
+    
+    def __repr__(self):
+        """
+        Returns a printable representation of this instance.
+        
+        @return: A descriptive string containing most of this instance fields
+        @rtype: str
+        """
+        return u"%s(element_id=%s, timestamp=%s, status=%s)" % (
+            self.__class__.__name__,
+            repr(self.element_id),
+            repr(self.timestamp),
+            repr(self.status),
+        )
+    
+    def __str__(self):
+        """
+        Coerces this instance to a string.
+        
+        @return: The status field
+        @rtype: str
+        """
+        return str(self.status)
 
 class Project(ORM_Base):
     """
@@ -503,33 +656,6 @@ class Project(ORM_Base):
         """
         return self._id
     
-    @classmethod
-    def query_tags(cls, session, tags={}):
-        """
-        Overriden query method to apply tag-based custom filtering, on top of
-        common equality filter.
-        
-        @param session: The database session in which to execute the query
-        @type session: Session
-        @param tags: Tag names and values to apply as a filter
-        @type tags: dict
-        @return: A query selecting Project instances, filtered by tags
-        @rtype: Query<L{Project}>
-        """
-        q = session.query(cls)
-        if tags:
-            crit = literal(False)
-            for tag, value in tags.iteritems():
-                if value == True: 
-                    crit |= (tables['tag_project'].c.tag == tag)
-                elif value == None or value != False :
-                    crit |= (tables['tag_project'].c.tag == tag) & (tables['tag_project'].c.value == value)
-            q = q.join(tables['tag_project']) \
-                 .group_by(*[c for c in tables['project'].columns]) \
-                 .having(func.count(tables['project'].c.id) == len(tags)) \
-                 .filter(crit)
-        return q.from_self()
-    
     def __repr__(self):
         """
         Returns a printable representation of this instance.
@@ -537,7 +663,11 @@ class Project(ORM_Base):
         @return: A descriptive string containing most of this instance fields
         @rtype: str
         """
-        return u"<Project %s: name '%s'>" % (self.id, self.name)
+        return u"%s(id=%s, name=%s)" % (
+            self.__class__.__name__,
+            repr(self.id),
+            repr(self.name),
+        )
     
     def __str__(self):
         """
@@ -548,7 +678,6 @@ class Project(ORM_Base):
         """
         return str(self.name)
 
-
 class TagElement(ORM_Base):
     """
     Mapping class for the table «tagelement».
@@ -557,17 +686,39 @@ class TagElement(ORM_Base):
     def __init__(self, *args, **kwargs):
         """
         Constructor for TagElement instances.
-        Ensures that the «tag» field is specified.
+        Ensures that the «name» field is specified.
         
         @param args: Optional arguments to the constructor
         @type args: tuple
         @param kwargs: Optional keyword arguments to the constructor
         @type kwargs: dict
         """
-        assert 'tag' in kwargs
-        assert kwargs['tag'] is not None
+        assert 'name' in kwargs
+        assert kwargs['name'] is not None
         super(TagElement, self).__init__(*args, **kwargs)
-
+    
+    def __repr__(self):
+        """
+        Returns a printable representation of this instance.
+        
+        @return: A descriptive string containing most of this instance fields
+        @rtype: str
+        """
+        return u"%s(element_id=%s, name=%s, value=%s)" % (
+            self.__class__.__name__,
+            repr(self.element_id),
+            repr(self.name),
+            repr(self.value),
+        )
+    
+    def __str__(self):
+        """
+        Coerces this instance to a string.
+        
+        @return: The name and value fields
+        @rtype: str
+        """
+        return str(dict(self.name, self.value))
 
 class TagProject(ORM_Base):
     """
@@ -577,86 +728,36 @@ class TagProject(ORM_Base):
     def __init__(self, *args, **kwargs):
         """
         Constructor for TagProject instances.
-        Ensures that the «tag» field is specified.
+        Ensures that the «name» field is specified.
         
         @param args: Optional arguments to the constructor
         @type args: tuple
         @param kwargs: Optional keyword arguments to the constructor
         @type kwargs: dict
         """
-        assert 'tag' in kwargs
-        assert kwargs['tag'] is not None
+        assert 'name' in kwargs
+        assert kwargs['name'] is not None
         super(TagProject, self).__init__(*args, **kwargs)
-
-
-# TODOS:
-#  - Reevaluate status cascading method. CTE? Triggers?
-
-if __name__ == '__main__': # pragma: no cover
     
-    config = json.load(open(os.path.dirname(__file__) + '/config.json', 'r'))
-    # Create and configure engine and session
-    engine = engine_from_config(config)
-    session_maker = scoped_session(sessionmaker(bind=engine, twophase=config['session.twophase']))
+    def __repr__(self):
+        """
+        Returns a printable representation of this instance.
+        
+        @return: A descriptive string containing most of this instance fields
+        @rtype: str
+        """
+        return u"%s(project_id=%s, name=%s, value=%s)" % (
+            self.__class__.__name__,
+            repr(self.project_id),
+            repr(self.name),
+            repr(self.value),
+        )
     
-    class EElement(Element):
-        pass
-    
-    class PProject(Project):
-        pass
-    
-    class TTagProject(TagProject):
-        pass
-    
-    class TTagElement(TagElement):
-        pass
-    
-    init(session_maker, extra_mapping={
-        'element'     : EElement,
-        'project'     : PProject,
-        'tag_project' : TTagProject,
-        'tag_element' : TTagElement,
-    })
-    
-    session = session_maker()
-    
-    metadata.drop_all()
-    metadata.create_all()
-    
-    p1 = PProject()
-    p1.name = 'p1'
-    
-    p2 = PProject(name='p2')
-    
-    p2.tag['t1'] = 't1'
-    p2.tag['t1'] = 'p2t1'
-    
-    p3 = PProject(name='p3')
-    
-    e1 = EElement(name='e1', status='50_DONE', forced_status=False, project=p1)
-    
-    e2 = EElement()
-    e2.name = 'e2'
-    e2.status = '30_PENDING'
-    e2.forced_status = True
-    e2.parent = e1
-    e2.project = p1
-    
-    e2.tag['t2'] = 't2'
-    e2.tag['t2'] = 'e2t2'
-
-    e3 = EElement(name='e3', status='05_ERROR', forced_status=False, project=p1)
-    
-    session.add(p1)
-    session.commit()
-    
-    e1.children.remove(e2)
-    e2.children.add(e1)
-    
-    p1.elements.remove(e3)
-    p2.elements.add(e3)
-    
-    session.commit()
-    
-    print session.query(PProject).filter_by(name='p1').one()
-    print session.query(EElement).filter_by(name='e1').one()
+    def __str__(self):
+        """
+        Coerces this instance to a string.
+        
+        @return: The name and value fields
+        @rtype: str
+        """
+        return str(dict(self.name, self.value))
